/**
    This program demonstrates how a "race condition" can come up
    when two threads both manipulate a shared variable. In this
    example, the shared variable is the class variable count.
    A race condition is when the result of a piece of code
    depends on the order in which the processor executes the
    threads.

    If you run this program on a multicore machine, you will see
    a fair number (but not a lot) of race conditions. However, if
    you run this program on a single core machine (or restrict the
    affinity of this program, and all of its threads, to a single
    core), then you will see very few race conditions. This shows
    how rare a race condition can be and why race conditions can be
    so hard to detect in real programs.
*/

public class Race_1v2
{
   // "global" variables
   public static long count  = 0;  // this variable is shared by thread1 and thread2
   public static long count1 = 0;  // this variable is used only by thread1
   public static long count2 = 0;  // this variable is used only by thread2

   public static void main(String[] args)
   {
      Function1 f1 = new Function1();
      Function2 f2 = new Function2();

      // create two threads
      System.out.println("Race_1v2: Starting threads");
      Thread t1 = new Thread( f1 );
      Thread t2 = new Thread( f2 );

      t1.start();
      t2.start();

      try
      {
         Thread.sleep(2000); // let the threads run for a while

         t1.interrupt();
         t2.interrupt();

         t1.join();
         t2.join();

         System.out.println("\nThreads stopped.");
         System.out.printf("count = %d, count1 = %d, count2 = %d\n", count, count1, count2);
      }
      catch (InterruptedException e)
      {
         e.getMessage();
      }
   }
}


class Function1 implements Runnable
{
   public void run()
   {
      long new_count, old_count;
      while (! Thread.interrupted())
      {
         // create a race situation with thread2
         // increment a global variable and store the results locally
         old_count = Race_1v2.count++;
         new_count = Race_1v2.count;
         // check that the local and global variables were consistent with each other
         if (new_count - old_count != 1)
            System.out.printf("Bingo1: new_count - old_count = %d\n", new_count-old_count);
         Race_1v2.count1++;
      }
      System.out.println("Thread 1 ending.");
   }
}


class Function2 implements Runnable
{
   public void run()
   {
      long new_count, old_count;
      while (! Thread.interrupted())
      {
         // create a race situation with thread1
         old_count = Race_1v2.count++;
         new_count = Race_1v2.count;
         // check for the race condition
         if (new_count - old_count != 1)
            System.out.printf("Bingo2: new_count - old_count = %d\n", new_count-old_count);
         Race_1v2.count2++;
      }
      System.out.println("Thread 2 ending.");
   }
}
